import enum
import json
import typing as T


class Policy(enum.Enum):
    none = "NONE"
    get = "GET_ONLY"
    read = "READ_ONLY"
    write = "WRITE_ONLY"
    read_write = "READ_WRITE"

    def bucket(
        self, bucket_name: str, *, json_encode: bool = True
    ) -> T.Union[str, T.Dict[str, T.Any]]:
        policies = {
            Policy.get: _get,
            Policy.read: _read,
            Policy.write: _write,
            Policy.read_write: _read_write,
            Policy.none: _none,
        }
        pol = policies[self](bucket_name)
        if json_encode:
            return json.dumps(pol)
        return pol


def _none(bucket_name: str) -> T.Dict:
    return {"Version": "2012-10-17", "Statement": []}


def _get(bucket_name: str) -> T.Dict:
    return {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": ["s3:GetObject"],
                "Resource": [f"arn:aws:s3:::{bucket_name}/*"],
            }
        ],
    }


def _read(bucket_name: str) -> T.Dict:
    return {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": ["s3:GetBucketLocation"],
                "Resource": [f"arn:aws:s3:::{bucket_name}"],
            },
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": ["s3:ListBucket"],
                "Resource": [f"arn:aws:s3:::{bucket_name}"],
            },
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": ["s3:GetObject"],
                "Resource": [f"arn:aws:s3:::{bucket_name}/*"],
            },
        ],
    }


def _write(bucket_name: str) -> T.Dict:
    return {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": ["s3:GetBucketLocation"],
                "Resource": [f"arn:aws:s3:::{bucket_name}"],
            },
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": ["s3:ListBucketMultipartUploads"],
                "Resource": [f"arn:aws:s3:::{bucket_name}"],
            },
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": [
                    "s3:ListMultipartUploadParts",
                    "s3:AbortMultipartUpload",
                    "s3:DeleteObject",
                    "s3:PutObject",
                ],
                "Resource": [f"arn:aws:s3:::{bucket_name}/*"],
            },
        ],
    }


def _read_write(bucket_name: str) -> T.Dict:
    return {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": ["s3:GetBucketLocation"],
                "Resource": [f"arn:aws:s3:::{bucket_name}"],
            },
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": ["s3:ListBucket"],
                "Resource": [f"arn:aws:s3:::{bucket_name}"],
            },
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": ["s3:ListBucketMultipartUploads"],
                "Resource": [f"arn:aws:s3:::{bucket_name}"],
            },
            {
                "Effect": "Allow",
                "Principal": {"AWS": ["*"]},
                "Action": [
                    "s3:AbortMultipartUpload",
                    "s3:DeleteObject",
                    "s3:GetObject",
                    "s3:ListMultipartUploadParts",
                    "s3:PutObject",
                ],
                "Resource": [f"arn:aws:s3:::{bucket_name}/*"],
            },
        ],
    }



from io import StringIO

import minio
from django.core.files.base import ContentFile
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase, override_settings
from minio_storage.policy import Policy

from .utils import BaseTestMixin, bucket_name


@override_settings(
    MINIO_STORAGE_MEDIA_USE_PRESIGNED=False, MINIO_STORAGE_STATIC_USE_PRESIGNED=False
)
class CommandsTests(BaseTestMixin, TestCase):
    @override_settings(
        MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET=False,
        MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY=False,
    )
    def test_management_command(self):

        storage = self.media_storage
        bucket = self.media_storage.bucket_name

        call_command("minio", "check")

        with self.assertRaisesRegex(CommandError, f"bucket {bucket} is not empty"):
            call_command("minio", "delete")

        try:
            self.obliterate_bucket(self.media_storage.bucket_name)
        except minio.error.NoSuchBucket:
            pass

        with self.assertRaisesRegex(CommandError, f"bucket {bucket} does not exist"):
            call_command("minio", "check")

        with self.assertRaisesRegex(CommandError, f"bucket {bucket} does not exist"):
            call_command("minio", "policy")

        with self.assertRaisesRegex(CommandError, f"bucket {bucket} does not exist"):
            call_command("minio", "ls")

        with self.assertRaisesRegex(CommandError, f"bucket {bucket} does not exist"):
            call_command("minio", "delete")

        call_command("minio", "create")

        call_command("minio", "check")

        with self.assertRaisesRegex(CommandError, f"you have already created {bucket}"):
            call_command("minio", "create")

        with self.assertRaisesRegex(CommandError, f"bucket {bucket} has no policy"):
            call_command("minio", "policy")

        for p in [p.value for p in Policy]:
            call_command("minio", "policy", "--set", p)

        call_command("minio", "policy", "--set", "GET_ONLY")

        call_command("minio", "policy")

        call_command("minio", "delete")

        with self.assertRaisesRegex(CommandError, f"bucket {bucket} does not exist"):
            call_command("minio", "check")

        call_command("minio", "create")

        files = [
            "animals/cats/cat1.txt",
            "animals/cats/cat2.txt",
            "animals/dogs/dog1.txt",
            "animals/dogs/dog2.txt",
            "what.txt",
        ]
        for p in files:
            self.assertEqual(p, storage.save(p, ContentFile(b"abc")))

        def ls_test(*args, expected):
            out = StringIO()
            call_command("minio", "ls", *args, stdout=out)
            out.seek(0)
            lines = out.read().splitlines()
            self.assertEqual(sorted(lines), sorted(expected))

        test_data = (
            (
                ["-r"],
                #
                [
                    "animals/dogs/dog1.txt",
                    "animals/dogs/dog2.txt",
                    "animals/cats/cat1.txt",
                    "animals/cats/cat2.txt",
                    "what.txt",
                ],
            ),
            (
                ["--files"],
                #
                ["what.txt"],
            ),
            (
                ["--dirs"],
                #
                ["animals/"],
            ),
            (
                [],
                #
                ["animals/", "what.txt"],
            ),
            (
                ["--prefix", "animals/"],
                #
                ["animals/dogs/", "animals/cats/"],
            ),
            (
                ["--prefix", "animals/", "--dirs"],
                #
                ["animals/dogs/", "animals/cats/"],
            ),
            (
                ["--prefix", "animals/", "--files"],
                #
                [],
            ),
            (
                ["-r", "-p", "animals/do"],
                #
                ["animals/dogs/dog1.txt", "animals/dogs/dog2.txt"],
            ),
            (
                ["-r", "-p", "animals/do", "--dirs"],
                #
                [],
            ),
        )

        for test in test_data:
            (args, expected) = test
            ls_test(*args, expected=expected)

    def test_check(self):
        out = StringIO()
        call_command("minio", "check", stdout=out)
        self.assertIn("", out.getvalue())

    def test_check_not_exists(self):
        name = bucket_name("new")
        out = StringIO()
        err = StringIO()
        with self.assertRaises(CommandError):
            call_command("minio", "--bucket", name, "check", stdout=out, stderr=err)
        self.assertEqual("", out.getvalue())
        self.assertEqual("", err.getvalue())

    def test_list_files(self):
        out = StringIO()
        call_command("minio", "ls", stdout=out)
        out.seek(0)
        lines = sorted(out.readlines())
        expected = sorted([f"{self.new_file}\n", f"{self.second_file}\n"])
        self.assertEqual(lines, expected)

    def test_list_buckets(self):
        out = StringIO()
        call_command("minio", "ls", "--buckets", stdout=out)
        out.seek(0)
        lines = sorted(out.readlines())
        self.assertIn(f"{self.media_storage.bucket_name}\n", lines)



import minio.error as merr


class MinIOError(OSError):
    def __init__(self, msg, cause):
        super().__init__(msg)
        self.cause = cause


reraise = {}
for v in (
    merr.APINotImplemented,
    merr.AccessDenied,
    merr.AccountProblem,
    merr.CredentialNotSupported,
    merr.CrossLocationLoggingProhibited,
    merr.ExpiredToken,
    merr.InvalidAccessKeyId,
    merr.InvalidAddressingHeader,
    merr.InvalidBucketError,
    merr.InvalidBucketName,
    merr.InvalidDigest,
    merr.InvalidEncryptionAlgorithmError,
    merr.InvalidEndpointError,
    merr.InvalidSecurity,
    merr.InvalidToken,
    merr.NoSuchBucket,
):
    reraise[v] = {"err": v}


def minio_error(msg, e):
    if e.__class__ in reraise:
        return e
    return MinIOError(msg, e)



import os

import requests
from django.conf import settings
from django.core.files.base import ContentFile, File
from django.test import TestCase, override_settings
from minio.error import InvalidAccessKeyId

from minio_storage.storage import MinioMediaStorage

from .utils import BaseTestMixin


@override_settings(
    MINIO_STORAGE_MEDIA_USE_PRESIGNED=True, MINIO_STORAGE_STATIC_USE_PRESIGNED=True
)
class UploadTests(BaseTestMixin, TestCase):
    def test_file_upload_success(self):
        self.media_storage.save("trivial.txt", ContentFile(b"12345"))

    @override_settings(
        MINIO_STORAGE_ACCESS_KEY="wrong_key", MINIO_STORAGE_SECRET_KEY="wrong_secret"
    )
    def test_file_upload_fail_incorrect_keys(self):
        with self.assertRaises(InvalidAccessKeyId):
            MinioMediaStorage()

    def test_two_files_with_the_same_name_can_be_uploaded(self):
        ivan = self.media_storage.save("pelican.txt", ContentFile(b"Ivan le Pelican"))
        jean = self.media_storage.save("pelican.txt", ContentFile(b"Jean le Pelican"))
        self.assertNotEqual(jean, ivan)

    def test_files_from_filesystem_are_uploaded_properly(self):
        f = File(open(os.path.join(settings.BASE_DIR, "watermelon-cat.jpg"), "br"))
        saved_file = self.media_storage.save("watermelon-cat.jpg", f)
        res = requests.get(self.media_storage.url(saved_file))
        self.assertAlmostEqual(
            round(res.content.__sizeof__() / 100), round(f.size / 100)
        )

    def test_files_are_uploaded_from_the_beginning(self):
        local_filename = os.path.join(settings.BASE_DIR, "watermelon-cat.jpg")
        f = open(local_filename, "br")
        f.seek(20000)
        saved_file = self.media_storage.save("watermelon-cat.jpg", f)
        file_size = os.stat(local_filename).st_size
        res = requests.get(self.media_storage.url(saved_file))
        self.assertAlmostEqual(
            round(res.content.__sizeof__() / 100), round(file_size / 100)
        )

    def test_files_cannot_be_open_in_write_mode(self):
        test_file = self.media_storage.save(
            "iomodetest.txt", ContentFile(b"should not change")
        )
        with self.assertRaises(NotImplementedError):
            self.media_storage.open(test_file, mode="bw")

    def test_files_seekable(self):
        self.media_storage.save("read_seek_test.txt", ContentFile(b"should not change"))

        f = self.media_storage.open("read_seek_test.txt", mode="br")
        f.seek(4)
        f.seek(0)

    def test_upload_and_get_back_file_with_funky_name(self):
        self.media_storage.save("áčďěščřžýŽŇůúť.txt", ContentFile(b"12345"))

    def test_uploaded_and_downloaded_file_sizes_match(self):
        pass

    def test_uploaded_files_end_up_in_the_right_bucket(self):
        pass

    def test_static_files_end_up_in_the_right_bucket(self):
        pass

    def test_upload_file_beggining_with_dot(self):
        self.media_storage.save(
            ".hidden_file", ContentFile(b"Not really, but whatever")
        )
        self.assertTrue(self.media_storage.exists(".hidden_file"))
        self.media_storage.delete(".hidden_file")
        self.assertFalse(self.media_storage.exists(".hidden_file"))

    def test_metadata(self):
        ivan = self.media_storage.save("pelican.txt", ContentFile(b"Ivan le Pelican"))
        res = self.media_storage.client.stat_object(
            self.media_storage.bucket_name, ivan
        )
        self.assertEqual(res.metadata, {"Content-Type": "text/plain"})


@override_settings(
    MINIO_STORAGE_MEDIA_OBJECT_METADATA={"Cache-Control": "max-age=1000"},
)
class TestDefaultObjectMetadata(BaseTestMixin, TestCase):
    def test_default_metadata(self):
        ivan = self.media_storage.save("pelican.txt", ContentFile(b"Ivan le Pelican"))
        res = self.media_storage.client.stat_object(
            self.media_storage.bucket_name, ivan
        )
        self.assertEqual(
            res.metadata,
            {"Cache-Control": "max-age=1000", "Content-Type": "text/plain"},
        )



import hashlib
import os
import warnings

from django.core.files.base import ContentFile
from minio import Minio
from minio_storage.storage import MinioMediaStorage, MinioStaticStorage, get_setting

warnings.simplefilter("default")
warnings.filterwarnings(
    "ignore", message="This usage is deprecated, please use pytest.* instead"
)


warnings.simplefilter("ignore", ResourceWarning)


def bucket_name(name):
    env_name = os.getenv("TOX_ENVNAME", "").encode("utf-8")
    env_hash = hashlib.md5(env_name).hexdigest()
    return "".join([name, env_hash])


class BaseTestMixin:
    @staticmethod
    def bucket_name(name):
        return bucket_name(name)

    def setUp(self):
        self.media_storage = MinioMediaStorage()
        self.static_storage = MinioStaticStorage()
        self.new_file = self.media_storage.save("test-file", ContentFile(b"yep"))
        self.second_file = self.media_storage.save("test-file", ContentFile(b"nope"))

    def tearDown(self):
        client = self.minio_client()
        self.obliterate_bucket(self.bucket_name("tests-media"), client=client)
        self.obliterate_bucket(self.bucket_name("tests-static"), client=client)

    def minio_client(self):
        minio_client = Minio(
            endpoint=get_setting("MINIO_STORAGE_ENDPOINT"),
            access_key=get_setting("MINIO_STORAGE_ACCESS_KEY"),
            secret_key=get_setting("MINIO_STORAGE_SECRET_KEY"),
            secure=get_setting("MINIO_STORAGE_USE_HTTPS"),
        )
        return minio_client

    def obliterate_bucket(self, name, client=None):
        if client is None:
            client = self.minio_client()

        for obj in client.list_objects(name, "", True):
            client.remove_object(name, obj.object_name)
        for obj in client.list_incomplete_uploads(name, ""):  # pragma: no cover  # noqa
            client.remove_incomplete_upload(name, obj.objectname)
        client.remove_bucket(name)



import tempfile
import typing as T
from logging import getLogger

from django.core.files.base import File
from minio import error as merr
from minio_storage.errors import minio_error

if T.TYPE_CHECKING:
    from minio_storage.storage import Storage

logger = getLogger("minio_storage")


class ReadOnlyMixin:
    """File class mixin which disallows .write() calls"""

    def writable(self) -> bool:
        return False

    def write(*args, **kwargs):
        raise NotImplementedError("this is a read only file")


class NonSeekableMixin:
    """File class mixin which disallows .seek() calls"""

    def seekable(self) -> bool:
        return False

    def seek(self, *args, **kwargs) -> bool:
        # TODO: maybe exception is better
        # raise NotImplementedError('seek is not supported')
        return False


class MinioStorageFile(File):
    def __init__(self, name: str, mode: str, storage: "Storage", **kwargs):
        self._storage: "Storage" = storage
        self.name: str = name
        self._mode: str = mode
        self._file = None


class ReadOnlyMinioObjectFile(MinioStorageFile, ReadOnlyMixin, NonSeekableMixin):
    """A django File class which directly exposes the underlying minio object. This
means the the instance doesnt support functions like .seek() and is required to
be closed to be able to reuse minio connections.

Note: This file class is not tested yet"""

    def __init__(
        self,
        name: str,
        mode: str,
        storage: "Storage",
        max_memory_size: T.Optional[int] = None,
        **kwargs,
    ):
        if mode.find("w") > -1:
            raise NotImplementedError(
                "ReadOnlyMinioObjectFile storage only support read modes"
            )
        if max_memory_size is not None:
            self.max_memory_size = max_memory_size
        super().__init__(name, mode, storage)

    def _get_file(self):
        if self._file is None:
            try:
                obj = self._storage.client.get_object(
                    self._storage.bucket_name, self.name
                )
                self._file = obj
                return self._file
            except merr.ResponseError as error:
                logger.warn(error)
                raise OSError(f"File {self.name} does not exist")
            finally:
                try:
                    obj.release_conn()
                except Exception as e:
                    logger.error(str(e))
        return self._file

    def _set_file(self, value):
        self._file = value

    file = property(_get_file, _set_file)

    def close(self):
        try:
            self.file.close()
        finally:
            self.file.release_conn()


class ReadOnlySpooledTemporaryFile(MinioStorageFile, ReadOnlyMixin):
    """A django File class which buffers the minio object into a local
SpooledTemporaryFile. """

    max_memory_size: int = 1024 * 1024 * 10

    def __init__(
        self,
        name: str,
        mode: str,
        storage: "Storage",
        max_memory_size: T.Optional[int] = None,
        **kwargs,
    ):
        if mode.find("w") > -1:
            raise NotImplementedError(
                "ReadOnlySpooledTemporaryFile storage only support read modes"
            )
        if max_memory_size is not None:
            self.max_memory_size = max_memory_size
        super().__init__(name, mode, storage)

    def _get_file(self):
        if self._file is None:
            try:
                obj = self._storage.client.get_object(
                    self._storage.bucket_name, self.name
                )
                self._file = tempfile.SpooledTemporaryFile(
                    max_size=self.max_memory_size
                )
                for d in obj.stream(amt=1024 * 1024):
                    self._file.write(d)
                self._file.seek(0)
                return self._file
            except merr.ResponseError as error:
                raise minio_error(f"File {self.name} does not exist", error)
            finally:
                try:
                    obj.release_conn()
                except Exception as e:
                    logger.error(str(e))
        return self._file

    def _set_file(self, value):
        self._file = value

    file = property(_get_file, _set_file)

    def close(self):
        if self._file is not None:
            self._file.close()
            self._file = None



import io
import os
import shutil
import tempfile

from django.core.files.base import ContentFile
from django.test import TestCase, override_settings
from django.utils.deconstruct import deconstructible
from minio_storage.files import ReadOnlyMinioObjectFile
from minio_storage.storage import (
    MinioStorage,
    create_minio_client_from_settings,
    get_setting,
)

from .utils import BaseTestMixin
from .utils import bucket_name as create_test_bucket_name


@deconstructible
class SecretStorage(MinioStorage):
    """The SecretStorage MinioStorage subclass can be used directly, as a storage in
    settings.DEFAULT_FILE_STORAGE or after instantiated used individually on any django
    FileField:

    from django.db import models

    ss = SecretStorage(bucket_name='invoices')

    class Invoice(models.Model):
        ...
        pdf = models.FileField(storage=ss)

    """

    # We can set a new default File class implementation that will be used here because
    # we want to stream the data directly from minio. Imagine that we need to process
    # large files where we don't want to waste time/ram/disk space on  writing the file
    # to disk two times before processing it.
    #
    file_class = ReadOnlyMinioObjectFile

    def __init__(self, bucket_name=None):

        # we can create the minio client ourselves or use
        # create_minio_client_from_settings convinience function while providing it with
        # extra args.
        #
        client = create_minio_client_from_settings(minio_kwargs={"region": "us-east-1"})

        # or use our own Django setting
        #
        if bucket_name is None:
            bucket_name = get_setting("SECRET_BUCKET_NAME")

        # Run the super constructor and make a choice to only use presigned urls with
        # this bucket so that we can keep files more private here than how media files
        # usually are public readable.
        #
        super().__init__(
            client,
            bucket_name,
            auto_create_bucket=True,
            auto_create_policy=False,
            presign_urls=True,
        )


class CustomStorageTests(BaseTestMixin, TestCase):
    @override_settings(SECRET_BUCKET_NAME=create_test_bucket_name("my-secret-bucket"))
    def test_custom_storage(self):
        # Instansiate a storage class and put a file in it so that we have something to
        # work with.
        #
        storage = SecretStorage()
        storage_filename = storage.save("secret.txt", ContentFile(b"abcd"))

        # Create a temporary workspace directory.
        #
        # It's importat that this directory is deleted after we are done so we use the
        # with statement here.
        #
        with tempfile.TemporaryDirectory() as workspace:

            # A filename to use for the file inside the working directory.
            #
            filename = os.path.join(workspace, "secret.txt")

            # Open a stream with the minio file objenct and the temporary file.
            #
            # We might be processing a lot of files in a loop here so we are going top
            # use the with statement to ensure that both the input stream and output
            # files are closed after the copying is done.
            #
            with open(filename, "wb") as out_file, storage.open(
                storage_filename
            ) as storage_file:

                # Copy the stream from the http stream to the out_file
                #
                shutil.copyfileobj(storage_file.file, out_file)

                #
                # We are not using the ReadOnlyMinioObjectFile type so we can't seek in
                # it.
                #
                with self.assertRaises(io.UnsupportedOperation):
                    storage_file.file.seek()

            workspace_files = os.listdir(workspace)
            print(workspace_files)  # prints: ['secret.txt']

            #
            # Process the file with external tools or something....
            #
            # For the purpouse of the example test we just check that the contents of
            # the file is what we wrote in the beginning of the test.
            #
            with open(filename, "rb") as f:
                self.assertEqual(f.read(), b"abcd")

        #
        # Clean up after the test
        #
        storage.delete(storage_filename)

        #
        # use the minio client directly to also remove bucket
        #
        storage.client.remove_bucket(storage.bucket_name)



import datetime
import mimetypes
import posixpath
import typing as T
import urllib
from logging import getLogger
from time import mktime
from urllib.parse import urlsplit, urlunsplit

import minio
import minio.error as merr
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.files.storage import Storage
from django.utils import timezone
from django.utils.deconstruct import deconstructible
from minio.helpers import get_target_url

from minio_storage.errors import minio_error
from minio_storage.files import ReadOnlySpooledTemporaryFile
from minio_storage.policy import Policy

logger = getLogger("minio_storage")


@deconstructible
class MinioStorage(Storage):
    """An implementation of Django's file storage using the minio client.

    The implementation should comply with
    https://docs.djangoproject.com/en/dev/ref/files/storage/.

    """

    file_class = ReadOnlySpooledTemporaryFile

    def __init__(
        self,
        minio_client: minio.Minio,
        bucket_name: str,
        *,
        base_url: T.Optional[str] = None,
        file_class=None,
        auto_create_bucket: bool = False,
        presign_urls: bool = False,
        auto_create_policy: bool = False,
        policy_type: T.Optional[Policy] = None,
        object_metadata: T.Optional[T.Dict[str, str]] = None,
        backup_format: T.Optional[str] = None,
        backup_bucket: T.Optional[str] = None,
        assume_bucket_exists: bool = False,
        **kwargs,
    ):
        self.client = minio_client
        self.bucket_name = bucket_name
        self.base_url = base_url

        self.backup_format = backup_format
        self.backup_bucket = backup_bucket
        if bool(self.backup_format) != bool(self.backup_bucket):
            raise ImproperlyConfigured(
                "To enable backups, make sure to set both backup format "
                "and backup format"
            )

        if file_class is not None:
            self.file_class = file_class
        self.auto_create_bucket = auto_create_bucket
        self.auto_create_policy = auto_create_policy
        self.assume_bucket_exists = assume_bucket_exists
        self.policy_type = policy_type
        self.presign_urls = presign_urls
        self.object_metadata = object_metadata

        self._init_check()

        # A base_url_client is only necessary when using presign_urls
        if self.presign_urls and self.base_url:
            # Do this after _init_check, so client's bucket region cache will
            # already be populated
            self.base_url_client = self._create_base_url_client(
                self.client, self.bucket_name, self.base_url
            )

        super().__init__()

    def _init_check(self):
        if not self.assume_bucket_exists:
            if self.auto_create_bucket and not self.client.bucket_exists(
                self.bucket_name
            ):
                self.client.make_bucket(self.bucket_name)
                if self.auto_create_policy:
                    policy_type = self.policy_type
                    if policy_type is None:
                        policy_type = Policy.get
                    self.client.set_bucket_policy(
                        self.bucket_name, policy_type.bucket(self.bucket_name)
                    )

            elif not self.client.bucket_exists(self.bucket_name):
                raise OSError(f"The bucket {self.bucket_name} does not exist")

    @staticmethod
    def _create_base_url_client(client: minio.Minio, bucket_name: str, base_url: str):
        """
        Clone a Minio client, using a different endpoint from `base_url`.
        """
        base_url_parts = urlsplit(base_url)

        # Clone from the normal client, but with base_url as the endpoint
        base_url_client = minio.Minio(
            base_url_parts.netloc,
            access_key=client._access_key,
            secret_key=client._secret_key,
            session_token=client._session_token,
            secure=base_url_parts.scheme == 'https',
            # The bucket region may be auto-detected by client (via an HTTP
            # request), so don't just use client._region
            region=client._get_bucket_region(bucket_name),
            http_client=client._http
        )
        if hasattr(client, '_credentials'):
            # Client credentials do not exist prior to minio-py 5.0.7, but
            # they should be reused if possible
            base_url_client._credentials = client._credentials

        return base_url_client

    def _sanitize_path(self, name):
        v = posixpath.normpath(name).replace("\\", "/")
        if v == ".":
            v = ""
        if name.endswith("/") and not v.endswith("/"):
            v += "/"
        return v

    def _examine_file(self, name, content):
        """Examines a file and produces information necessary for upload.

        Returns a tuple of the form (content_size, content_type,
        sanitized_name)

        """
        content_size = content.size
        content_type = mimetypes.guess_type(name, strict=False)
        content_type = content_type[0] or "application/octet-stream"
        sane_name = self._sanitize_path(name)
        return (content_size, content_type, sane_name)

    def _open(self, name, mode="rb"):
        try:
            f = self.file_class(self._sanitize_path(name), mode, self)
        except merr.MinioError as e:
            raise minio_error("File {} could not be saved: {}".format(name, str(e)), e)
        return f

    def _save(self, name: str, content: bytes) -> str:
        try:
            if hasattr(content, "seek") and callable(content.seek):
                content.seek(0)
            content_size, content_type, sane_name = self._examine_file(name, content)
            self.client.put_object(
                self.bucket_name,
                sane_name,
                content,
                content_size,
                content_type,
                metadata=self.object_metadata,
            )
            return sane_name
        except merr.ResponseError as error:
            raise minio_error(f"File {name} could not be saved", error)

    def delete(self, name: str) -> None:
        if self.backup_format and self.backup_bucket:
            try:
                obj = self.client.get_object(self.bucket_name, name)
            except merr.ResponseError as error:
                raise minio_error(
                    "Could not obtain file {} " "to make a copy of it".format(name),
                    error,
                )

            try:
                content_length = int(obj.getheader("Content-Length"))
            except ValueError as error:
                raise minio_error(f"Could not backup removed file {name}", error)

            # Creates the backup filename
            target_name = "{}{}".format(
                timezone.now().strftime(self.backup_format), name
            )
            try:
                self.client.put_object(
                    self.backup_bucket, target_name, obj, content_length
                )
            except merr.ResponseError as error:
                raise minio_error(
                    "Could not make a copy of file "
                    "{} before removing it".format(name),
                    error,
                )

        try:
            self.client.remove_object(self.bucket_name, name)
        except merr.ResponseError as error:
            raise minio_error(f"Could not remove file {name}", error)

    def exists(self, name: str) -> bool:
        try:
            self.client.stat_object(self.bucket_name, self._sanitize_path(name))
            return True
        except merr.ResponseError as error:
            # TODO - deprecate
            if error.code == "NoSuchKey":
                return False
            else:
                raise minio_error(f"Could not stat file {name}", error)
        except merr.NoSuchKey:
            return False
        except merr.NoSuchBucket:
            raise
        except Exception as error:
            logger.error(error)

    def listdir(self, path: str) -> T.Tuple[T.List, T.List]:
        #  [None, "", "."] is supported to mean the configured root among various
        #  implementations of Storage implementations so we copy that behaviour even if
        #  maybe None should raise an exception instead.
        #
        #  If the path prefix does not match anything full prefix that does exist this
        #  function will just return empty results, this is different from
        #  FileSystemStorage where an invalid directory would raise an OSError.

        if path in [None, "", ".", "/"]:
            path = ""
        else:
            if not path.endswith("/"):
                path += "/"

        dirs: T.List[str] = []
        files: T.List[str] = []
        try:
            objects = self.client.list_objects_v2(self.bucket_name, prefix=path)
            for o in objects:
                p = posixpath.relpath(o.object_name, path)
                if o.is_dir:
                    dirs.append(p)
                else:
                    files.append(p)
            return dirs, files
        except merr.NoSuchBucket:
            raise
        except merr.ResponseError as error:
            raise minio_error(f"Could not list directory {path}", error)

    def size(self, name: str) -> int:
        try:
            info = self.client.stat_object(self.bucket_name, name)
            return info.size
        except merr.ResponseError as error:
            raise minio_error(f"Could not access file size for {name}", error)

    def _presigned_url(
        self, name: str, max_age: T.Optional[datetime.timedelta] = None
    ) -> str:
        kwargs = {}
        if max_age is not None:
            kwargs["expires"] = max_age

        client = (
            self.client
            if self.base_url is None else
            self.base_url_client
        )
        url = client.presigned_get_object(self.bucket_name, name, **kwargs)

        if self.base_url is not None:
            url_parts = urlsplit(url)
            base_url_parts = urlsplit(self.base_url)

            # It's assumed that self.base_url will contain bucket information,
            # which could be different, so remove the bucket_name component (with 1
            # extra character for the leading "/") from the generated URL
            url_key_path = url_parts.path[len(self.bucket_name) + 1:]

            # Prefix the URL with any path content from base_url
            new_url_path = base_url_parts.path + url_key_path

            # Reconstruct the URL with an updated path
            url = urlunsplit((
                url_parts.scheme,
                url_parts.netloc,
                new_url_path,
                url_parts.query,
                url_parts.fragment
            ))
        return url

    def url(
        self, name: str, *args, max_age: T.Optional[datetime.timedelta] = None
    ) -> str:
        if self.presign_urls:
            url = self._presigned_url(name, max_age=max_age)
        else:
            if self.base_url is not None:

                def strip_beg(path):
                    while path.startswith("/"):
                        path = path[1:]
                    return path

                def strip_end(path):
                    while path.endswith("/"):
                        path = path[:-1]
                    return path

                url = "{}/{}".format(
                    strip_end(self.base_url), urllib.parse.quote(strip_beg(name))
                )
            else:
                url = get_target_url(
                    self.client._endpoint_url,
                    bucket_name=self.bucket_name,
                    object_name=name,
                    # bucket_region=region,
                )
        return url

    def accessed_time(self, name: str) -> datetime.datetime:
        """
        Not available via the S3 API
        """
        return self.modified_time(name)

    def created_time(self, name: str) -> datetime.datetime:
        """
        Not available via the S3 API
        """
        return self.modified_time(name)

    def modified_time(self, name: str) -> datetime.datetime:
        try:
            info = self.client.stat_object(self.bucket_name, name)
            return datetime.datetime.fromtimestamp(mktime(info.last_modified))
        except merr.ResponseError as error:
            raise minio_error(
                f"Could not access modification time for file {name}", error
            )


_NoValue = object()


def get_setting(name, default=_NoValue):
    result = getattr(settings, name, default)
    if result is _NoValue:
        # print("Attr {} : {}".format(name, getattr(settings, name, default)))
        raise ImproperlyConfigured
    else:
        return result


def create_minio_client_from_settings(*, minio_kwargs=dict()):
    endpoint = get_setting("MINIO_STORAGE_ENDPOINT")
    access_key = get_setting("MINIO_STORAGE_ACCESS_KEY")
    secret_key = get_setting("MINIO_STORAGE_SECRET_KEY")
    secure = get_setting("MINIO_STORAGE_USE_HTTPS", True)
    client = minio.Minio(
        endpoint,
        access_key=access_key,
        secret_key=secret_key,
        secure=secure,
        **minio_kwargs,
    )
    return client


@deconstructible
class MinioMediaStorage(MinioStorage):
    def __init__(self):
        client = create_minio_client_from_settings()
        bucket_name = get_setting("MINIO_STORAGE_MEDIA_BUCKET_NAME")
        base_url = get_setting("MINIO_STORAGE_MEDIA_URL", None)
        auto_create_bucket = get_setting(
            "MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET", False
        )
        auto_create_policy = get_setting(
            "MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY", "GET_ONLY"
        )

        policy_type = Policy.get
        if isinstance(auto_create_policy, str):
            policy_type = Policy(auto_create_policy)
            auto_create_policy = True

        presign_urls = get_setting("MINIO_STORAGE_MEDIA_USE_PRESIGNED", False)
        backup_format = get_setting("MINIO_STORAGE_MEDIA_BACKUP_FORMAT", False)
        backup_bucket = get_setting("MINIO_STORAGE_MEDIA_BACKUP_BUCKET", False)

        assume_bucket_exists = get_setting(
            "MINIO_STORAGE_ASSUME_MEDIA_BUCKET_EXISTS", False
        )

        object_metadata = get_setting("MINIO_STORAGE_MEDIA_OBJECT_METADATA", None)
        # print("SETTING", object_metadata)

        super().__init__(
            client,
            bucket_name,
            auto_create_bucket=auto_create_bucket,
            auto_create_policy=auto_create_policy,
            policy_type=policy_type,
            base_url=base_url,
            presign_urls=presign_urls,
            backup_format=backup_format,
            backup_bucket=backup_bucket,
            assume_bucket_exists=assume_bucket_exists,
            object_metadata=object_metadata,
        )


@deconstructible
class MinioStaticStorage(MinioStorage):
    def __init__(self):
        client = create_minio_client_from_settings()
        base_url = get_setting("MINIO_STORAGE_STATIC_URL", None)
        bucket_name = get_setting("MINIO_STORAGE_STATIC_BUCKET_NAME")
        auto_create_bucket = get_setting(
            "MINIO_STORAGE_AUTO_CREATE_STATIC_BUCKET", False
        )
        auto_create_policy = get_setting(
            "MINIO_STORAGE_AUTO_CREATE_STATIC_POLICY", "GET_ONLY"
        )

        policy_type = Policy.get
        if isinstance(auto_create_policy, str):
            policy_type = Policy(auto_create_policy)
            auto_create_policy = True

        presign_urls = get_setting("MINIO_STORAGE_STATIC_USE_PRESIGNED", False)

        assume_bucket_exists = get_setting(
            "MINIO_STORAGE_ASSUME_STATIC_BUCKET_EXISTS", False
        )

        object_metadata = get_setting("MINIO_STORAGE_STATIC_OBJECT_METADATA", None)

        super().__init__(
            client,
            bucket_name,
            auto_create_bucket=auto_create_bucket,
            auto_create_policy=auto_create_policy,
            policy_type=policy_type,
            base_url=base_url,
            presign_urls=presign_urls,
            assume_bucket_exists=assume_bucket_exists,
            object_metadata=object_metadata,
        )



import datetime
import io
import os

import requests
from freezegun import freeze_time

from django.core.files.base import ContentFile
from django.test import TestCase, override_settings
from minio.error import NoSuchKey
from minio_storage.errors import MinIOError
from minio_storage.storage import MinioMediaStorage
import unittest
from .utils import BaseTestMixin


@override_settings(
    MINIO_STORAGE_MEDIA_USE_PRESIGNED=True, MINIO_STORAGE_STATIC_USE_PRESIGNED=True
)
class RetrieveTestsWithRestrictedBucket(BaseTestMixin, TestCase):
    def test_presigned_url_generation(self):
        media_test_file_name = self.media_storage.save(
            "weird & ÜRΛ", ContentFile(b"irrelevant")
        )
        url = self.media_storage.url(media_test_file_name)
        res = requests.get(url)
        self.assertEqual(res.content, b"irrelevant")

        static_test_file_name = self.static_storage.save(
            "weird & ÜRΛ", ContentFile(b"irrelevant")
        )
        url = self.static_storage.url(static_test_file_name)
        res = requests.get(url)
        self.assertEqual(res.content, b"irrelevant")

    def test_url_of_non_existent_object(self):
        self.media_storage.url("this does not exist")

    def test_file_size(self):
        test_file = self.media_storage.save("sizetest.txt", ContentFile(b"1234"))
        self.assertEqual(4, self.media_storage.size(test_file))

    def test_size_of_non_existent_throws(self):
        test_file = self.media_storage.save("sizetest.txt", ContentFile(b"1234"))
        self.media_storage.delete(test_file)
        with self.assertRaises(NoSuchKey):
            self.media_storage.size(test_file)

    def test_modified_time(self):
        self.assertIsInstance(
            self.media_storage.modified_time(self.new_file), datetime.datetime
        )

    def test_accessed_time(self):
        self.assertIsInstance(
            self.media_storage.accessed_time(self.new_file), datetime.datetime
        )

    def test_created_time(self):
        self.assertIsInstance(
            self.media_storage.created_time(self.new_file), datetime.datetime
        )

    def test_modified_time_of_non_existent_throws(self):
        with self.assertRaises(NoSuchKey):
            self.media_storage.modified_time("nonexistent.jpg")

    def _listdir_root(self, root):
        self.media_storage.save("dir1/file2.txt", ContentFile(b"meh"))
        test_dir = self.media_storage.listdir(root)
        (dirs, files) = test_dir
        self.assertEqual(dirs, ["dir1"])
        self.assertEqual(files, sorted([self.new_file, self.second_file]))
        self.assertEqual(len(files), 2)
        self.assertEqual(len(test_dir), 2)

    def test_listdir_emptystr(self):
        self._listdir_root("")

    def test_listdir_dot(self):
        self._listdir_root(".")

    def test_listdir_none(self):
        self._listdir_root(None)

    def test_listdir_slash(self):
        self._listdir_root("/")

    def _listdir_sub(self, path):
        self.media_storage.save("dir/file.txt", ContentFile(b"meh"))
        self.media_storage.save("dir/file2.txt", ContentFile(b"meh"))
        self.media_storage.save("dir/dir3/file3.txt", ContentFile(b"meh"))
        dirs, files = self.media_storage.listdir("dir/")
        self.assertEqual((dirs, files), (["dir3"], ["file.txt", "file2.txt"]))

    def test_listdir_subdir_slash(self):
        self._listdir_sub("dir/")

    def test_listdir_subdir_noslash(self):
        self._listdir_sub("dir")

    def test_list_nonexist(self):
        test_dir = self.media_storage.listdir("nonexist")
        self.assertEqual(test_dir, ([], []))

    def test_list_prefix(self):
        self.media_storage.save("dir/file.txt", ContentFile(b"meh"))
        test_dir = self.media_storage.listdir("di")
        self.assertEqual(test_dir, ([], []))

    def test_file_exists(self):
        existent = self.media_storage.save("existent.txt", ContentFile(b"meh"))
        self.assertTrue(self.media_storage.exists(existent))

    def test_file_exists_failure(self):
        self.assertFalse(self.media_storage.exists("nonexistent.txt"))

    @unittest.skip("Skipping this test because undecided if it should raise exception")
    def test_opening_non_existing_file_raises_oserror(self):
        with self.assertRaises(OSError):
            self.media_storage.open("this does not exist")

    @unittest.skip("Skipping this test because undecided if it should raise exception")
    def test_opening_non_existing_file_raises_minioerror(self):
        with self.assertRaises(MinIOError):
            self.media_storage.open("this does not exist")
        try:
            self.media_storage.open("this does not exist")
        except MinIOError as e:
            assert e.cause.__class__ == NoSuchKey

    def test_file_names_are_properly_sanitized(self):
        self.media_storage.save("./meh22222.txt", io.BytesIO(b"stuff"))

    def test_url_max_age(self):
        url = self.media_storage.url(
            "test-file", max_age=datetime.timedelta(seconds=10)
        )
        self.assertEqual(requests.get(url).status_code, 200)

    def test_max_age_too_old(self):
        with freeze_time(-datetime.timedelta(seconds=10)):
            url = self.media_storage.url(
                "test-file", max_age=datetime.timedelta(seconds=10)
            )
        self.assertEqual(requests.get(url).status_code, 403)

    def test_no_file(self):
        url = self.media_storage.url("no-file", max_age=datetime.timedelta(seconds=10))
        self.assertEqual(requests.get(url).status_code, 404)

    def test_max_age_no_file(self):
        with freeze_time(-datetime.timedelta(seconds=10)):
            url = self.media_storage.url(
                "no-file", max_age=datetime.timedelta(seconds=10)
            )
        self.assertEqual(requests.get(url).status_code, 403)


class URLTests(TestCase):
    @override_settings(
        MINIO_STORAGE_MEDIA_USE_PRESIGNED=False,
        MINIO_STORAGE_MEDIA_URL=None,
        MINIO_STORAGE_MEDIA_BUCKET_NAME="foo",
    )
    def test_no_base_url(self):
        endpoint = os.getenv("MINIO_STORAGE_ENDPOINT", "minio:9000")
        assert endpoint != ""
        media_storage = MinioMediaStorage()
        url = media_storage.url("22")
        self.assertEqual(url, f"http://{endpoint}/foo/22")

    @override_settings(
        MINIO_STORAGE_MEDIA_USE_PRESIGNED=False,
        MINIO_STORAGE_MEDIA_URL=None,
        MINIO_STORAGE_MEDIA_BUCKET_NAME="foo",
    )
    def test_no_base_url_subpath(self):
        endpoint = os.getenv("MINIO_STORAGE_ENDPOINT", "minio:9000")
        assert endpoint != ""
        media_storage = MinioMediaStorage()
        name = "23/23/aaa/bbb/22"
        url = media_storage.url(name)
        self.assertEqual(url, f"http://{endpoint}/foo/23/23/aaa/bbb/22")

    @override_settings(
        MINIO_STORAGE_MEDIA_USE_PRESIGNED=True,
        MINIO_STORAGE_MEDIA_URL=None,
        MINIO_STORAGE_MEDIA_BUCKET_NAME="foo",
    )
    def test_presigned_no_base_url(self):
        endpoint = os.getenv("MINIO_STORAGE_ENDPOINT", "minio:9000")
        assert endpoint != ""
        media_storage = MinioMediaStorage()
        url = media_storage.url("22")
        self.assertRegex(url, rf"^http://{endpoint}/foo/22\?")

    @override_settings(
        MINIO_STORAGE_MEDIA_USE_PRESIGNED=False,
        MINIO_STORAGE_MEDIA_URL="https://example23.com/foo",
        MINIO_STORAGE_MEDIA_BUCKET_NAME="bar",
    )
    def test_base_url(self):
        media_storage = MinioMediaStorage()
        url = media_storage.url("1")
        self.assertEqual(url, "https://example23.com/foo/1")

    @override_settings(
        MINIO_STORAGE_MEDIA_USE_PRESIGNED=False,
        MINIO_STORAGE_MEDIA_URL="https://example23.com/foo",
        MINIO_STORAGE_MEDIA_BUCKET_NAME="bar",
    )
    def test_base_url_subpath(self):
        media_storage = MinioMediaStorage()
        url = media_storage.url("1/2/3/4")
        self.assertEqual(url, "https://example23.com/foo/1/2/3/4")

    @override_settings(
        MINIO_STORAGE_MEDIA_URL="http://example11.com/foo",
        MINIO_STORAGE_MEDIA_USE_PRESIGNED=True,
        MINIO_STORAGE_MEDIA_BUCKET_NAME="bar",
    )
    def test_presigned_base_url(self):
        media_storage = MinioMediaStorage()
        url = media_storage.url("1")
        self.assertIn("X-Amz-Signature", url)
        self.assertRegex(url, r"^http://example11.com/foo/1\?")

    @override_settings(
        MINIO_STORAGE_MEDIA_URL="http://example11.com/foo",
        MINIO_STORAGE_MEDIA_USE_PRESIGNED=True,
        MINIO_STORAGE_MEDIA_BUCKET_NAME="bar",
    )
    def test_presigned_base_url_subpath(self):
        media_storage = MinioMediaStorage()
        url = media_storage.url("1/555/666/777")
        self.assertIn("X-Amz-Signature", url)
        self.assertRegex(url, r"^http://example11.com/foo/1/555/666/777\?")

    BASE = "http://base/url"
    NAME = "Bö/ &öl@:/E"
    ENCODED = "B%C3%B6/%20%26%C3%B6l%40%3A/E"

    @override_settings(
        MINIO_STORAGE_MEDIA_URL=None,
        MINIO_STORAGE_MEDIA_USE_PRESIGNED=False,
        MINIO_STORAGE_MEDIA_BUCKET_NAME="encoding",
    )
    def test_quote_url(self):
        media_storage = MinioMediaStorage()
        url = media_storage.url(self.NAME)
        self.assertTrue(url.endswith(self.ENCODED))
        self.assertTrue(len(url) > len(self.ENCODED))

    @override_settings(
        MINIO_STORAGE_MEDIA_URL=BASE,
        MINIO_STORAGE_MEDIA_USE_PRESIGNED=False,
        MINIO_STORAGE_MEDIA_BUCKET_NAME="encoding",
    )
    def test_quote_base_url(self):
        media_storage = MinioMediaStorage()
        url = media_storage.url(self.NAME)
        self.assertEqual(url, f"{self.BASE}/{self.ENCODED}")

    @override_settings(
        MINIO_STORAGE_MEDIA_URL=BASE,
        MINIO_STORAGE_MEDIA_USE_PRESIGNED=True,
        MINIO_STORAGE_MEDIA_BUCKET_NAME="encoding",
    )
    def test_quote_base_url_presigned(self):
        media_storage = MinioMediaStorage()
        url = media_storage.url(self.NAME)
        prefix = f"{self.BASE}/{self.ENCODED}"
        self.assertTrue(url.startswith(prefix))
        self.assertTrue(len(url) > len(prefix))



import json
import sys
from string import Template
from unittest.mock import patch

import minio.error
from django.core.management.base import BaseCommand, CommandError
from django.utils.module_loading import import_string
from minio_storage.policy import Policy
from minio_storage.storage import MinioStorage
import argparse


class Command(BaseCommand):
    help = "verify, list, create and delete minio buckets"

    CHECK = "check"
    CREATE = "create"
    DELETE = "delete"
    LIST = "ls"
    POLICY = "policy"

    FULL_FORMAT = "$name $size $modified $url $etag"

    def add_arguments(self, parser):

        group = parser.add_argument_group("minio")
        group.add_argument(
            "--class",
            type=str,
            default="minio_storage.storage.MinioMediaStorage",
            help="Storage class to modify "
            "(media/static are short names for default classes)",
        )
        group.add_argument(
            "--bucket",
            type=str,
            default=None,
            help="bucket name (default: storage defined bucket if not set)",
        )

        cmds = parser.add_subparsers(
            dest="command",
            title="subcommands",
            description="valid subcommands",
            # required=True,
        )
        cmds._parser_class = argparse.ArgumentParser  # circumvent Django 1.11 bug

        cmds.add_parser(self.CHECK, help="check bucket")

        cmds.add_parser(self.CREATE, help="make bucket")

        cmds.add_parser(self.DELETE, help="remove an empty bucket")

        ls = cmds.add_parser(self.LIST, help="list bucket objects or buckets")
        ls.add_argument("--dirs", action="store_true", help="include directories")
        ls.add_argument("--files", action="store_true", help="include files")
        ls.add_argument(
            "-r", "--recursive", action="store_true", help="find files recursive"
        )
        ls.add_argument("-p", "--prefix", type=str, default="", help="path prefix")
        ls.add_argument(
            "--buckets", action="store_true", help="list buckets instead of files"
        )
        ls.add_argument(
            "-f",
            "--format",
            type=str,
            default="$name",
            help="list format. ( $name $size $modified $url $etag )",
        )

        policy = cmds.add_parser(self.POLICY, help="get or set bucket policy")
        policy.add_argument(
            "--set",
            type=str,
            default=None,
            choices=[p.value for p in Policy],
            help="set bucket policy",
        )

        super().add_arguments(parser)

    def handle(self, *args, **options):
        storage = self.storage(options)
        bucket_name = options["bucket"] or storage.bucket_name
        command = options["command"] or ""
        if command == self.CHECK:
            return self.bucket_exists(storage, bucket_name)
        if command == self.CREATE:
            return self.bucket_create(storage, bucket_name)
        elif command == self.DELETE:
            return self.bucket_delete(storage, bucket_name)
        elif command == self.LIST:
            if options["buckets"]:
                return self.list_buckets(storage)

            list_dirs = True
            list_files = True
            summary = True
            if options["dirs"] or options["files"]:
                list_dirs = options["dirs"]
                list_files = options["files"]
                summary = False

            return self.bucket_list(
                storage,
                bucket_name,
                prefix=options["prefix"],
                list_dirs=list_dirs,
                list_files=list_files,
                recursive=options["recursive"],
                format=options["format"],
                summary=summary,
            )
        elif command == self.POLICY:
            if options["set"] is not None:
                return self.policy_set(
                    storage, bucket_name, policy=Policy(options["set"])
                )
            return self.policy_get(storage, bucket_name)
        self.print_help("minio", "")
        if command != "":
            raise CommandError(f"don't know how to handle command: {command}")
        raise CommandError("command name required")

    def storage(self, options):
        class_name = {
            "media": "minio_storage.storage.MinioMediaStorage",
            "static": "minio_storage.storage.MinioStaticStorage",
        }.get(options["class"], options["class"])

        try:
            storage_class = import_string(class_name)
        except ImportError:
            raise CommandError(f"could not find storage class: {class_name}")
        if not issubclass(storage_class, MinioStorage):
            raise CommandError(f"{class_name} is not an sub class of MinioStorage.")

        # TODO: maybe another way
        with patch.object(storage_class, "_init_check", return_value=None):
            storage = storage_class()
            return storage

    def bucket_exists(self, storage, bucket_name):
        exists = storage.client.bucket_exists(bucket_name)
        if not exists:
            raise CommandError(f"bucket {bucket_name} does not exist")

    def list_buckets(self, storage):
        objs = storage.client.list_buckets()
        for o in objs:
            self.stdout.write(f"{o.name}")

    def bucket_list(
        self,
        storage,
        bucket_name: str,
        *,
        prefix: str,
        list_dirs: bool,
        list_files: bool,
        recursive: bool,
        format: str = None,
        summary: bool = True,
    ):
        try:
            objs = storage.client.list_objects_v2(
                bucket_name, prefix=prefix, recursive=recursive
            )

            template = None
            if format is not None and format != "$name":
                template = Template(format)

            def fmt(o):
                if template is None:
                    return o.object_name
                return template.substitute(
                    name=o.object_name,
                    size=o.size,
                    modified=o.last_modified,
                    etag=o.etag,
                    url=storage.url(o.object_name),
                )

            n_files = 0
            n_dirs = 0
            for o in objs:
                if o.is_dir:
                    n_dirs += 1
                    if list_dirs:
                        self.stdout.write(fmt(o))
                else:
                    n_files += 1
                    if list_files:
                        self.stdout.write(fmt(o))

            if summary:
                print(f"{n_files} files and {n_dirs} directories", file=sys.stderr)
        except minio.error.NoSuchBucket:
            raise CommandError(f"bucket {bucket_name} does not exist")

    def bucket_create(self, storage, bucket_name):
        try:
            storage.client.make_bucket(bucket_name)
            print(f"created bucket: {bucket_name}", file=sys.stderr)
        except minio.error.BucketAlreadyOwnedByYou:
            raise CommandError(f"you have already created {bucket_name}")
        return

    def bucket_delete(self, storage, bucket_name):
        try:
            storage.client.remove_bucket(bucket_name)
        except minio.error.NoSuchBucket:
            raise CommandError(f"bucket {bucket_name} does not exist")
        except minio.error.BucketNotEmpty:
            raise CommandError(f"bucket {bucket_name} is not empty")

    def policy_get(self, storage, bucket_name):
        try:
            policy = storage.client.get_bucket_policy(bucket_name)
            policy = json.loads(policy)
            policy = json.dumps(policy, ensure_ascii=False, indent=2)
            return policy
        except minio.error.NoSuchBucket:
            raise CommandError(f"bucket {bucket_name} does not exist")
        except minio.error.NoSuchBucketPolicy:
            raise CommandError(f"bucket {bucket_name} has no policy")

    def policy_set(self, storage, bucket_name, policy: Policy):
        try:
            policy = Policy(policy)
            storage.client.set_bucket_policy(bucket_name, policy.bucket(bucket_name))
        except minio.error.NoSuchBucket as e:
            raise CommandError(e.message)



# encoding: utf-8
from setuptools import setup

setup(
    name="django-minio-storage",
    license="MIT",
    use_scm_version=True,
    description="Django file storage using the minio python client",
    author="Tom Houlé",
    author_email="tom@kafunsho.be",
    url="https://github.com/py-pa/django-minio-storage",
    packages=[
        "minio_storage",
        "minio_storage/management/",
        "minio_storage/management/commands/",
    ],
    setup_requires=["setuptools_scm"],
    install_requires=["django>=1.11", "minio>=4.0.21"],
    extras_require={"test": ["coverage", "requests"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Framework :: Django",
    ],
)



import json

import requests

import minio
from django.core.exceptions import ImproperlyConfigured
from django.core.files.base import ContentFile
from django.test import TestCase, override_settings
from minio_storage.policy import Policy
from minio_storage.storage import MinioMediaStorage, MinioStaticStorage, get_setting

from .utils import BaseTestMixin


class BucketTests(BaseTestMixin, TestCase):
    @override_settings(
        MINIO_STORAGE_MEDIA_BUCKET_NAME="inexistent",
        MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET=False,
    )
    def test_media_storage_cannot_be_initialized_without_bucket(self):
        with self.assertRaises(OSError):
            MinioMediaStorage()

    @override_settings(
        MINIO_STORAGE_STATIC_BUCKET_NAME="inexistent",
        MINIO_STORAGE_AUTO_CREATE_STATIC_BUCKET=False,
    )
    def test_static_storage_cannot_be_initialized_without_bucket(self):
        with self.assertRaises(OSError):
            MinioStaticStorage()

    def test_get_setting_throws_early(self):
        with self.assertRaises(ImproperlyConfigured):
            get_setting("INEXISTENT_SETTING")

    @override_settings(
        MINIO_STORAGE_MEDIA_BUCKET_NAME="inexistent",
        MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET=False,
        MINIO_STORAGE_ASSUME_MEDIA_BUCKET_EXISTS=True,
    )
    def test_media_storage_ignore_bucket_check(self):
        MinioMediaStorage()

    @override_settings(
        MINIO_STORAGE_STATIC_BUCKET_NAME="inexistent",
        MINIO_STORAGE_AUTO_CREATE_STATIC_BUCKET=False,
        MINIO_STORAGE_ASSUME_STATIC_BUCKET_EXISTS=True,
    )
    def test_static_storage_ignore_bucket_check(self):
        MinioStaticStorage()


class BucketPolicyTests(BaseTestMixin, TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        self.obliterate_bucket(self.bucket_name("tests-media"))

    def assertPolicyEqual(self, first, second):
        """A stupid method to compare policies which is enough for this test"""

        def comparable(v):
            return sorted(json.dumps(v))

        def pretty(v):
            return json.dumps(v, sort_keys=True, indent=2)

        if comparable(first) != comparable(second):
            raise ValueError(
                f"not equal:\n\n ------{pretty(first)}\n\n ------{pretty(second)}"
            )

    @override_settings(
        MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET=True,
        MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY=False,
    )
    def test_auto_create_no_policy(self):
        ms = MinioMediaStorage()
        with self.assertRaises(minio.error.NoSuchBucketPolicy):
            ms.client.get_bucket_policy(ms.bucket_name)

    @override_settings(
        MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY=True,
        MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET=True,
    )
    def test_media_policy_auto_true(self):
        ms = MinioMediaStorage()
        self.maxDiff = 50000
        self.assertPolicyEqual(
            Policy.get.bucket(ms.bucket_name, json_encode=False),
            json.loads(ms.client.get_bucket_policy(ms.bucket_name)),
        )
        fn = ms.save("somefile", ContentFile(b"test"))
        self.assertEqual(ms.open(fn).read(), b"test")
        url = ms.url(fn)
        self.assertEqual(requests.get(url).status_code, 200)
        self.assertEqual(
            requests.get(f"{ms.client._endpoint_url}/{ms.bucket_name}").status_code, 403
        )

    @override_settings(
        MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY="GET_ONLY",
        MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET=True,
    )
    def test_media_policy_get(self):
        ms = MinioMediaStorage()
        self.maxDiff = 50000
        self.assertPolicyEqual(
            Policy.get.bucket(ms.bucket_name, json_encode=False),
            json.loads(ms.client.get_bucket_policy(ms.bucket_name)),
        )
        fn = ms.save("somefile", ContentFile(b"test"))
        self.assertEqual(ms.open(fn).read(), b"test")
        self.assertEqual(requests.get(ms.url(fn)).status_code, 200)
        self.assertEqual(
            requests.get(f"{ms.client._endpoint_url}/{ms.bucket_name}").status_code, 403
        )

    @override_settings(
        MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY="WRITE_ONLY",
        MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET=True,
    )
    def test_media_policy_write(self):
        ms = MinioMediaStorage()
        self.maxDiff = 50000
        self.assertPolicyEqual(
            Policy.write.bucket(ms.bucket_name, json_encode=False),
            json.loads(ms.client.get_bucket_policy(ms.bucket_name)),
        )
        fn = ms.save("somefile", ContentFile(b"test"))
        self.assertEqual(ms.open(fn).read(), b"test")
        self.assertEqual(requests.get(ms.url(fn)).status_code, 403)
        self.assertEqual(
            requests.get(f"{ms.client._endpoint_url}/{ms.bucket_name}").status_code, 403
        )

    @override_settings(
        MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY="READ_WRITE",
        MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET=True,
    )
    def test_media_policy_rw(self):
        ms = MinioMediaStorage()
        self.maxDiff = 50000
        self.assertPolicyEqual(
            Policy.read_write.bucket(ms.bucket_name, json_encode=False),
            json.loads(ms.client.get_bucket_policy(ms.bucket_name)),
        )

    @override_settings(
        MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY=True,
        MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET=True,
        MINIO_STORAGE_AUTO_CREATE_STATIC_POLICY=True,
        MINIO_STORAGE_AUTO_CREATE_STATIC_BUCKET=True,
    )
    def test_public_url_generation(self):
        media_storage = MinioMediaStorage()
        media_test_file_name = media_storage.save(
            "weird & ÜRΛ", ContentFile(b"irrelevant")
        )
        url = media_storage.url(media_test_file_name)
        res = requests.get(url)
        self.assertEqual(res.content, b"irrelevant")

        static_storage = MinioStaticStorage()
        static_test_file_name = static_storage.save(
            "weird & ÜRΛ", ContentFile(b"irrelevant")
        )
        url = static_storage.url(static_test_file_name)
        res = requests.get(url)
        self.assertEqual(res.content, b"irrelevant")
